@@ -12,7 +12,6 @@
 @todo Expanded copy/paste (Unusure how far to go)
 	@li Copy formula, value, serialized, mathml, latex?
 	@li Paste serialized, value?
-@bug Has the same Maemo tab color bug as DialCentral
 
 Some useful things on Maemo
 @li http://maemo.org/api_refs/4.1/libosso-2.16-1/group__Statesave.html
@@ -27,15 +26,13 @@ import sys
 import gc
 import os
 import string
+import logging
 import warnings
 
 import gtk
 import gtk.glade
 
-try:
-	import hildon
-except ImportError:
-	hildon = None
+import hildonize
 
 from libraries import gtkpie
 from libraries import gtkpieboard
@@ -46,10 +43,14 @@ import gtk_toolbox
 import constants
 
 
+_moduleLogger = logging.getLogger("ejpi_glade")
+
 PLUGIN_SEARCH_PATHS = [
 	os.path.join(os.path.dirname(__file__), "plugins/"),
 ]
 
+PROFILE_STARTUP = False
+
 
 class ValueEntry(object):
 
@@ -113,6 +114,8 @@ class Calculator(object):
 	_user_settings = "%s/settings.ini" % _user_data
 	_user_history = "%s/history.stack" % _user_data
 
+	MIN_BUTTON_SIZE = min(800, 480) // 6 - 20
+
 	def __init__(self):
 		self.__constantPlugins = plugin_utils.ConstantPluginManager()
 		self.__constantPlugins.add_path(*self._plugin_search_paths)
@@ -134,7 +137,7 @@ class Calculator(object):
 
 		self.__keyboardPlugins = plugin_utils.KeyboardPluginManager()
 		self.__keyboardPlugins.add_path(*self._plugin_search_paths)
-		self.__activeKeyboards = {}
+		self.__activeKeyboards = []
 
 		for path in self._glade_files:
 			if os.path.isfile(path):
@@ -155,29 +158,26 @@ class Calculator(object):
 
 		self._app = None
 		self._isFullScreen = False
-		if hildon is not None:
-			self._app = hildon.Program()
-			oldWindow = self._window
-			self._window = hildon.Window()
-			oldWindow.get_child().reparent(self._window)
-			self._app.add_window(self._window)
-			hildon.hildon_helper_set_thumb_scrollbar(self._widgetTree.get_widget('scrollingHistory'), True)
+		self._app = hildonize.get_app_class()()
+		self._window = hildonize.hildonize_window(self._app, self._window)
 
-			gtkMenu = self._widgetTree.get_widget("mainMenubar")
-			menu = gtk.Menu()
-			for child in gtkMenu.get_children():
-				child.reparent(menu)
-			self._window.set_menu(menu)
-			gtkMenu.destroy()
+		menu = hildonize.hildonize_menu(
+			self._window,
+			self._widgetTree.get_widget("mainMenubar"),
+			[]
+		)
 
-			self._window.connect("key-press-event", self._on_key_press)
-			self._window.connect("window-state-event", self._on_window_state_change)
-		else:
-			pass # warnings.warn("No Hildon", UserWarning, 2)
+		for scrollingWidgetName in (
+			"scrollingHistory",
+		):
+			scrollingWidget = self._widgetTree.get_widget(scrollingWidgetName)
+			assert scrollingWidget is not None, scrollingWidgetName
+			hildonize.hildonize_scrollwindow_with_viewport(scrollingWidget)
 
 		self.__errorDisplay = gtk_toolbox.ErrorDisplay(self._widgetTree)
 		self.__userEntry = ValueEntry(self._widgetTree.get_widget("entryView"))
 		self.__stackView = self._widgetTree.get_widget("historyView")
+		self.__pluginButton = self._widgetTree.get_widget("keyboardSelectionButton")
 
 		self.__historyStore = gtkhistory.GtkCalcHistory(self.__stackView)
 		self.__history = history.RpnCalcHistory(
@@ -187,6 +187,7 @@ class Calculator(object):
 		)
 		self.__load_history()
 
+		# Basic keyboard stuff
 		self.__sliceStyle = gtkpie.generate_pie_style(gtk.Button())
 		self.__handler = gtkpieboard.KeyboardHandler(self._on_entry_direct)
 		self.__handler.register_command_handler("push", self._on_push)
@@ -194,15 +195,22 @@ class Calculator(object):
 		self.__handler.register_command_handler("backspace", self._on_entry_backspace)
 		self.__handler.register_command_handler("clear", self._on_entry_clear)
 
+		# Main keyboard
 		builtinKeyboardId = self.__keyboardPlugins.lookup_plugin("Builtin")
 		self.__keyboardPlugins.enable_plugin(builtinKeyboardId)
 		self.__builtinPlugin = self.__keyboardPlugins.keyboards["Builtin"].construct_keyboard()
 		self.__builtinKeyboard = self.__builtinPlugin.setup(self.__history, self.__sliceStyle, self.__handler)
-		self._widgetTree.get_widget("functionLayout").pack_start(self.__builtinKeyboard)
+		self._widgetTree.get_widget("mainKeyboard").pack_start(self.__builtinKeyboard)
+		for child in self.__builtinKeyboard.get_children():
+			child.set_size_request(self.MIN_BUTTON_SIZE, self.MIN_BUTTON_SIZE)
+
+		# Plugins
 		self.enable_plugin(self.__keyboardPlugins.lookup_plugin("Trigonometry"))
 		self.enable_plugin(self.__keyboardPlugins.lookup_plugin("Computer"))
 		self.enable_plugin(self.__keyboardPlugins.lookup_plugin("Alphabet"))
+		self._set_plugin_kb(0)
 
+		# Callbacks
 		callbackMapping = {
 			"on_calculator_quit": self._on_close,
 			"on_paste": self._on_paste,
@@ -212,12 +220,18 @@ class Calculator(object):
 		self._widgetTree.signal_autoconnect(callbackMapping)
 		self._widgetTree.get_widget("copyMenuItem").connect("activate", self._on_copy)
 		self._widgetTree.get_widget("copyEquationMenuItem").connect("activate", self._on_copy_equation)
+		self._window.connect("key-press-event", self._on_key_press)
+		self._window.connect("window-state-event", self._on_window_state_change)
+		self._widgetTree.get_widget("entryView").connect("activate", self._on_push)
+		self.__pluginButton.connect("clicked", self._on_kb_plugin_selection_button)
 
-		if hildon is None:
-			self._window.set_title("%s" % constants.__pretty_app_name__)
+		hildonize.set_application_title(self._window, "%s" % constants.__pretty_app_name__)
 		self._window.connect("destroy", self._on_close)
 		self._window.show_all()
 
+		if not hildonize.IS_HILDON_SUPPORTED:
+			_moduleLogger.warning("No hildonization support")
+
 		try:
 			import osso
 		except ImportError:
@@ -228,7 +242,7 @@ class Calculator(object):
 			device = osso.DeviceState(self._osso)
 			device.set_device_state_callback(self._on_device_state_change, 0)
 		else:
-			pass # warnings.warn("No OSSO", UserWarning, 2)
+			_moduleLogger.warning("No OSSO support")
 
 	def display_error_message(self, msg):
 		error_dialog = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, msg)
@@ -245,16 +259,38 @@ class Calculator(object):
 		pluginName = pluginData[0]
 		plugin = self.__keyboardPlugins.keyboards[pluginName].construct_keyboard()
 		pluginKeyboard = plugin.setup(self.__history, self.__sliceStyle, self.__handler)
+		for child in pluginKeyboard.get_children():
+			child.set_size_request(self.MIN_BUTTON_SIZE, self.MIN_BUTTON_SIZE)
 
-		keyboardTabs = self._widgetTree.get_widget("pluginKeyboards")
-		keyboardTabs.append_page(pluginKeyboard, gtk.Label(pluginName))
-		keyboardPageNum = keyboardTabs.page_num(pluginKeyboard)
-		assert keyboardPageNum not in self.__activeKeyboards
-		self.__activeKeyboards[keyboardPageNum] = {
+		self.__activeKeyboards.append({
 			"pluginName": pluginName,
 			"plugin": plugin,
 			"pluginKeyboard": pluginKeyboard,
-		}
+		})
+
+	def _on_kb_plugin_selection_button(self, *args):
+		try:
+			pluginNames = [plugin["pluginName"] for plugin in self.__activeKeyboards]
+			oldIndex = pluginNames.index(self.__pluginButton.get_label())
+			newIndex = hildonize.touch_selector(self._window, "Keyboards", pluginNames, oldIndex)
+			self._set_plugin_kb(newIndex)
+		except Exception:
+			self.__errorDisplay.push_exception()
+
+	def _set_plugin_kb(self, pluginIndex):
+		plugin = self.__activeKeyboards[pluginIndex]
+		self.__pluginButton.set_label(plugin["pluginName"])
+
+		pluginParent = self._widgetTree.get_widget("pluginKeyboard")
+		oldPluginChildren = pluginParent.get_children()
+		if oldPluginChildren:
+			assert len(oldPluginChildren) == 1, "%r" % (oldPluginChildren, )
+			pluginParent.remove(oldPluginChildren[0])
+			oldPluginChildren[0].hide()
+		pluginKeyboard = plugin["pluginKeyboard"]
+		pluginParent.pack_start(pluginKeyboard)
+
+		pluginKeyboard.show_all()
 
 	def __load_history(self):
 		serialized = []
@@ -282,36 +318,45 @@ class Calculator(object):
 
 		@note Hildon specific
 		"""
-		if memory_low:
-			gc.collect()
+		try:
+			if memory_low:
+				gc.collect()
 
-		if save_unsaved_data or shutdown:
-			self.__save_history()
+			if save_unsaved_data or shutdown:
+				self.__save_history()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_window_state_change(self, widget, event, *args):
 		"""
 		@note Hildon specific
 		"""
-		if event.new_window_state & gtk.gdk.WINDOW_STATE_FULLSCREEN:
-			self._isFullScreen = True
-		else:
-			self._isFullScreen = False
+		try:
+			if event.new_window_state & gtk.gdk.WINDOW_STATE_FULLSCREEN:
+				self._isFullScreen = True
+			else:
+				self._isFullScreen = False
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_close(self, *args, **kwds):
-		if self._osso is not None:
-			self._osso.close()
-
 		try:
-			self.__save_history()
-		finally:
-			gtk.main_quit()
+			if self._osso is not None:
+				self._osso.close()
+
+			try:
+				self.__save_history()
+			finally:
+				gtk.main_quit()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_copy(self, *args):
 		try:
 			equationNode = self.__history.history.peek()
 			result = str(equationNode.evaluate())
 			self._clipboard.set_text(result)
-		except StandardError, e:
+		except Exception:
 			self.__errorDisplay.push_exception()
 
 	def _on_copy_equation(self, *args):
@@ -319,42 +364,70 @@ class Calculator(object):
 			equationNode = self.__history.history.peek()
 			equation = str(equationNode)
 			self._clipboard.set_text(equation)
-		except StandardError, e:
+		except Exception:
 			self.__errorDisplay.push_exception()
 
 	def _on_paste(self, *args):
-		contents = self._clipboard.wait_for_text()
-		self.__userEntry.append(contents)
+		try:
+			contents = self._clipboard.wait_for_text()
+			self.__userEntry.append(contents)
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_key_press(self, widget, event, *args):
 		"""
 		@note Hildon specific
 		"""
-		if event.keyval == gtk.keysyms.F6:
-			if self._isFullScreen:
-				self._window.unfullscreen()
-			else:
-				self._window.fullscreen()
+		try:
+			RETURN_TYPES = (gtk.keysyms.Return, gtk.keysyms.ISO_Enter, gtk.keysyms.KP_Enter)
+			if (
+				event.keyval == gtk.keysyms.F6 or
+				event.keyval in RETURN_TYPES and event.get_state() & gtk.gdk.CONTROL_MASK
+			):
+				if self._isFullScreen:
+					self._window.unfullscreen()
+				else:
+					self._window.fullscreen()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_push(self, *args):
-		self.__history.push_entry()
+		try:
+			self.__history.push_entry()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_unpush(self, *args):
-		self.__historyStore.unpush()
+		try:
+			self.__historyStore.unpush()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_entry_direct(self, keys, modifiers):
-		if "shift" in modifiers:
-			keys = keys.upper()
-		self.__userEntry.append(keys)
+		try:
+			if "shift" in modifiers:
+				keys = keys.upper()
+			self.__userEntry.append(keys)
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_entry_backspace(self, *args):
-		self.__userEntry.pop()
+		try:
+			self.__userEntry.pop()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_entry_clear(self, *args):
-		self.__userEntry.clear()
+		try:
+			self.__userEntry.clear()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_clear_all(self, *args):
-		self.__history.clear()
+		try:
+			self.__history.clear()
+		except Exception:
+			self.__errorDisplay.push_exception()
 
 	def _on_about_activate(self, *args):
 		dlg = gtk.AboutDialog()
@@ -389,10 +462,9 @@ def run_calculator():
 	gtk.gdk.threads_init()
 
 	gtkpie.IMAGES.add_path(os.path.join(os.path.dirname(__file__), "libraries/images"), )
-	if hildon is not None:
-		gtk.set_application_name(constants.__pretty_app_name__)
 	handle = Calculator()
-	gtk.main()
+	if not PROFILE_STARTUP:
+		gtk.main()
 
 
 class DummyOptions(object):
@@ -402,6 +474,7 @@ class DummyOptions(object):
 
 
 if __name__ == "__main__":
+	logging.basicConfig(level=logging.DEBUG)
 	if len(sys.argv) > 1:
 		try:
 			import optparse
